

import numpy as np
from scipy.optimize import brentq


def zero_crossings(f, x):
    """Get root of function f in intervall x"""
    y = f(x)
    zix = np.where(np.abs(np.diff(np.sign(y))) == 2)[0]

    x0 = np.array([brentq(f, x[i], x[i+1]) for i in zix])
    # y0 = np.array([f(i) for i in x0])

    return x0 #, y0 # the function values y0 should be ~0


def extrema(x, y=None):
    """Get extrema of curve x"""
    zix = np.where(np.abs(np.diff(np.sign(np.diff(x)))) == 2)[0]
    return zix
    if not y:
        print(zix)



import numpy as np
import xobjects as xo

from ._pyparticles import Pyparticles

from ..general import _pkg_root

from scipy.constants import m_p
from scipy.constants import e as qe
from scipy.constants import c as clight

pmass = m_p * clight * clight / qe

LAST_INVALID_STATE = -999999999

size_vars = (
    (xo.Int64,   '_capacity'),
    (xo.Int64,   '_num_active_particles'),
    (xo.Int64,   '_num_lost_particles'),
    )
# Capacity is always kept up to date
# the other two are placeholders to be used if needed
# i.e. on ContextCpu

scalar_vars = (
    (xo.Float64, 'q0'),
    (xo.Float64, 'mass0'),
    )

part_energy_vars = (
    (xo.Float64, 'psigma'),
    (xo.Float64, 'delta'),
    (xo.Float64, 'rpp'),
    (xo.Float64, 'rvv'),
        )

per_particle_vars = (
    (
        (xo.Float64, 'p0c'),
        (xo.Float64, 'gamma0'),
        (xo.Float64, 'beta0'),
        (xo.Float64, 's'),
        (xo.Float64, 'x'),
        (xo.Float64, 'y'),
        (xo.Float64, 'px'),
        (xo.Float64, 'py'),
        (xo.Float64, 'zeta'),
    )
    + part_energy_vars +
    (
        (xo.Float64, 'chi'),
        (xo.Float64, 'charge_ratio'),
        (xo.Float64, 'weight'),
        (xo.Int64, 'particle_id'),
        (xo.Int64, 'at_element'),
        (xo.Int64, 'at_turn'),
        (xo.Int64, 'state'),
        (xo.Int64, 'parent_particle_id'),
        (xo.UInt32, '__rng_s1'),
        (xo.UInt32, '__rng_s2'),
        (xo.UInt32, '__rng_s3'),
        (xo.UInt32, '__rng_s4')
    )
    )


fields = {}
for tt, nn in size_vars + scalar_vars:
    fields[nn] = tt

for tt, nn in per_particle_vars:
    fields[nn] = tt[:]

ParticlesData = type(
        'ParticlesData',
        (xo.Struct,),
        fields)

ParticlesData.extra_sources = [
    _pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
    _pkg_root.joinpath('random_number_generator/rng_src/particles_rng.h')]
ParticlesData.custom_kernels = {
    'Particles_initialize_rand_gen': xo.Kernel(
        args=[
            xo.Arg(ParticlesData, name='particles'),
            xo.Arg(xo.UInt32, pointer=True, name='seeds'),
            xo.Arg(xo.Int32, name='n_init')],
        n_threads='n_init')}



class Particles(xo.dress(ParticlesData, rename={
                             'delta': '_delta',
                             'psigma': '_psigma',
                             'rvv': '_rvv',
                             'rpp': '_rpp'})):

    """
        Particle objects have the following fields:

             - s [m]:  Reference accumulated pathlength
             - x [m]:  Horizontal position
             - px[1]:  Px / (m/m0 * p0c) = beta_x gamma /(beta0 gamma0)
             - y [m]:  Vertical position
             - py [1]:  Py / (m/m0 * p0c)
             - delta[1]:  Pc / (m/m0 * p0c) - 1
             - ptau [1]:  Energy / (m/m0 * p0c) - 1
             - psigma [1]:  ptau/beta0
             - rvv [1]:  beta/beta0
             - rpp [1]:  1/(1+delta) = (m/m0 * p0c) / Pc
             - zeta [m]:  beta (s/beta0 - ct )
             - tau [m]:
             - sigma [m]:  s - beta0 ct = rvv * zeta
             - mass0 [eV]:
             - q0 [e]:  Reference charge
             - p0c [eV]: Reference momentum
             - energy0 [eV]: Reference energy
             - gamma0 [1]:  Reference relativistic gamma
             - beta0 [1]:  Reference relativistix beta
             - chi [1]:  q/ q0 * m0/m = qratio / mratio
             - mass_ratio [1]:  mass/mass0
             - charge_ratio [1]:  q / q0
             - particle_id [int]: Identifier of the particle
             - at_turn [int]:  Number of tracked turns
             - state [int]:  It is ``0`` if the particle is lost, ``1`` otherwise
             - weight [int]:  Particle weight in number of particles
                              (for collective sims.)
             - at_element [int]: Identifier of the last element through which
                                 the particle has been
             - parent_particle_id [int]: Identifier of the parent particle
                                         (secondary production processes)
    """

    _structure = {
            'size_vars': size_vars,
            'scalar_vars': scalar_vars,
            'per_particle_vars': per_particle_vars}

    def to_dict(self, copy_to_cpu=True):
        if copy_to_cpu:
            cpobj = self.copy(_context=xo.context_default)
            return cpobj.to_dict(copy_to_cpu=False)
        else:
            dct = super().to_dict()
            dct['delta'] = self.delta
            dct['psigma'] = self.psigma
            dct['rvv'] = self.rvv
            dct['rpp'] = self.rpp
        return dct

    def to_pandas(self):
        import pandas as pd
        return pd.DataFrame(self.to_dict())

    @classmethod
    def from_pandas(cls, df, _context=None, _buffer=None, _offset=None):
        dct = df.to_dict(orient='list')
        for tt, nn in scalar_vars + size_vars:
            if nn in dct.keys() and not np.isscalar(dct[nn]):
                dct[nn] = dct[nn][0]
        return cls(**dct, _context=_context, _buffer=_buffer, _offset=_offset)

    @classmethod
    def merge(cls, lst, _context=None, _buffer=None, _offset=None):

        # TODO For now the merge is performed on CPU for add contexts.
        # Slow for objects on GPU (transferred to CPU for the merge).

        # Move everything to cpu
        cpu_lst = []
        for pp in lst:
            assert isinstance(pp, cls)
            if isinstance(pp._buffer.context, xo.ContextCpu):
                cpu_lst.append(pp)
            else:
                cpu_lst.append(pp.copy(_context=xo.context_default))

        # Check that scalar variable are compatible
        for tt, nn in scalar_vars:
            vals = [getattr(pp, nn) for pp in cpu_lst]
            assert np.allclose(vals,
                                getattr(cpu_lst[0], nn), rtol=0, atol=1e-14)

        # Make new particle on CPU
        capacity = np.sum([pp._capacity for pp in cpu_lst])
        new_part_cpu = cls(_capacity=capacity)

        # Copy scalar vars from first particle
        for tt, nn in scalar_vars:
            setattr(new_part_cpu, nn, getattr(cpu_lst[0], nn))

        # Copy per-particle vars
        first = 0
        max_id_curr = -1
        for pp in cpu_lst:
            for tt, nn in per_particle_vars:
                if not(nn == 'particle_id' or nn == 'parent_id'):
                    getattr(new_part_cpu, nn)[
                            first:first+pp._capacity] = getattr(pp, nn)

            # Handle particle_ids and parent_ids
            mask = pp.particle_id >= 0
            new_id = pp.particle_id.copy()
            new_parent_id = pp.parent_particle_id.copy()
            if np.min(new_id[mask]) <= max_id_curr:
                new_id[mask] += (max_id_curr + 1)
                new_parent_id[mask] += (max_id_curr + 1)
            new_part_cpu.particle_id[first:first+len(new_id)] = new_id
            new_part_cpu.parent_particle_id[
                    first:first+len(new_id)] = new_parent_id

            max_id_curr = np.max(new_id)
            first += pp._capacity

        # Reorganize
        new_part_cpu.reorganize()

        # Copy to appropriate context
        if _context is None and _buffer is None:
            # Use constext of first particle
            if isinstance(lst[0]._buffer.context, xo.ContextCpu):
                new_part_cpu._buffer.context = lst[0]._buffer.context
                return new_part_cpu
            else:
                return new_part_cpu.copy(_context=lst[0]._buffer.context)
        else:
            return new_part_cpu.copy(_context=_context, _buffer=_buffer,
                                     _offset=_offset)

    def filter(self, mask):

        if isinstance(self._buffer.context, xo.ContextCpu):
            self_cpu = self
        else:
            self_cpu = self.copy(_context=xo.context_default)

        # copy mask to cpu is needed
        if isinstance(mask, self._buffer.context.nplike_array_type):
            mask = self._buffer.context.nparray_from_context_array(mask)

            # Pyopencl returns int8 instead of bool
            if (isinstance(self._buffer.context, xo.ContextPyopencl) and
                mask.dtype==np.int8):
                assert np.all((mask>=0) & (mask<=1))
                mask = mask > 0

        # Make new particle on CPU
        test_x = self_cpu.x[mask]
        capacity = len(test_x)
        new_part_cpu = self.__class__(_capacity=capacity)

        # Copy scalar vars from first particle
        for tt, nn in scalar_vars:
            setattr(new_part_cpu, nn, getattr(self_cpu, nn))

        # Copy per-particle vars
        for tt, nn in per_particle_vars:
            getattr(new_part_cpu, nn)[:] = getattr(self_cpu, nn)[mask]

        # Reorganize
        new_part_cpu.reorganize()

        # Copy to original context 
        target_ctx = self._buffer.context
        if isinstance(target_ctx, xo.ContextCpu):
            new_part_cpu._buffer.context = target_ctx
            return new_part_cpu
        else:
            return new_part_cpu.copy(_context=target_ctx)

    def __init__(self, **kwargs):

        input_kwargs = kwargs.copy()

        if '_xobject' in kwargs.keys():
            # Initialize xobject
            self.xoinitialize(**kwargs)
        else:
            if any([nn in kwargs.keys() for tt, nn in per_particle_vars]):
                # Needed to generate consistent longitudinal variables
                pyparticles = Pyparticles(**kwargs)

                part_dict = _pyparticles_to_xpart_dict(pyparticles)
                if ('_capacity' in kwargs.keys() and
                         kwargs['_capacity'] is not None):
                    assert kwargs['_capacity'] >= part_dict['_capacity']
                else:
                    kwargs['_capacity'] = part_dict['_capacity']
            else:
                pyparticles = None
                if '_capacity' not in kwargs.keys():
                    kwargs['_capacity'] = 1

            # Make sure _capacity is integer
            kwargs['_capacity'] = int(kwargs['_capacity'])

            # We just provide array sizes to xoinitialize (we will set values later)
            kwargs.update(
                    {kk: kwargs['_capacity'] for tt, kk in per_particle_vars})

            # Initialize xobject
            self.xoinitialize(**kwargs)

            # Initialize coordinates
            if pyparticles is not None:
                context = self._buffer.context
                for tt, kk in list(scalar_vars):
                    setattr(self, kk, part_dict[kk])
                for tt, kk in list(per_particle_vars):
                    if kk.startswith('__'):
                        continue
                    vv = getattr(self, kk)
                    vals =  context.nparray_to_context_array(part_dict[kk])
                    ll = len(vals)
                    vv[:ll] = vals
                    vv[ll:] = LAST_INVALID_STATE
            else:
                for tt, kk in list(scalar_vars):
                    setattr(self, kk, 0.)

                for tt, kk in list(per_particle_vars):
                    if kk == 'chi' or kk == 'charge_ratio' or kk == 'state':
                        value = 1.
                    elif kk == 'particle_id':
                        value = np.arange(0, self._capacity, dtype=np.int64)
                    else:
                        value = 0.
                    getattr(self, kk)[:] = value

        self._num_active_particles = -1 # To be filled in only on CPU
        self._num_lost_particles = -1 # To be filled in only on CPU

        # Force values provided by user if compatible
        for nn in part_energy_varnames():
            vvv = self._buffer.context.nparray_from_context_array(getattr(self, nn))
            if nn in input_kwargs.keys():
                if np.allclose(vvv, input_kwargs[nn], rtol=0, atol=1e-13):
                    if np.isscalar(input_kwargs[nn]):
                        getattr(self, "_"+nn)[:] = input_kwargs[nn]
                    else:
                        getattr(self, "_"+nn)[:] = (
                                context.nparray_to_context_array(
                                    np.array(input_kwargs[nn])))
                else:
                    raise ValueError(
                            'What?! This should have been intercepted before!')

        if isinstance(self._buffer.context, xo.ContextCpu):
            # Particles always need to be organized to run on CPU
            self.reorganize()

    def _init_random_number_generator(self, seeds=None):

         self.compile_custom_kernels(only_if_needed=True)

         if seeds is None:
            seeds = np.random.randint(low=1, high=4e9,
                        size=self._capacity, dtype=np.uint32)
         else:
            assert len(seeds) == particles._capacity
            if not hasattr(seeds, 'dtype') or seeds.dtype != np.uint32:
                seeds = np.array(seeds, dtype=np.uint32)

         context = self._buffer.context
         seeds_dev = context.nparray_to_context_array(seeds)
         context.kernels.Particles_initialize_rand_gen(particles=self,
             seeds=seeds_dev, n_init=self._capacity)

    def hide_lost_particles(self):
         self._lim_arrays_name = '_num_active_particles'

    def unhide_lost_particles(self):
         del(self._lim_arrays_name)

    @property
    def lost_particles_are_hidden(self):
         return (hasattr(self, '_lim_arrays_name') and
                 self._lim_arrays_name == '_num_active_particles')

    def reorganize(self):
        assert not isinstance(self._buffer.context, xo.ContextPyopencl), (
                'Masking does not work with pyopencl')

        if self.lost_particles_are_hidden:
            restore_hidden = True
            self.unhide_lost_particles()
        else:
            restore_hidden = False

        mask_active = self.state > 0
        mask_lost = (self.state < 1) & (self.state>LAST_INVALID_STATE)

        n_active = np.sum(mask_active)
        n_lost = np.sum(mask_lost)

        for tt, nn in self._structure['per_particle_vars']:
            vv = getattr(self, nn)
            vv_active = vv[mask_active]
            vv_lost = vv[mask_lost]

            vv[:n_active] = vv_active
            vv[n_active:n_active+n_lost] = vv_lost
            vv[n_active+n_lost:] = LAST_INVALID_STATE

        if isinstance(self._buffer.context, xo.ContextCpu):
            self._num_active_particles = n_active
            self._num_lost_particles = n_lost

        if restore_hidden:
            self.hide_lost_particles()

        return n_active, n_lost

    def add_particles(self, part, keep_lost=False):

        if keep_lost:
            raise NotImplementedError
        assert not isinstance(self._buffer.context, xo.ContextPyopencl), (
                'Masking does not work with pyopencl')

        mask_copy = part.state > 0
        n_copy = np.sum(mask_copy)

        n_active, n_lost = self.reorganize()
        i_start_copy = n_active + n_lost
        n_free = self._capacity - n_active - n_lost

        max_id = np.max(self.particle_id[:n_active+n_lost])

        if n_copy > n_free:
            raise NotImplementedError("Out of space, need to regenerate xobject")

        for tt, nn in self._structure['scalar_vars']:
            assert np.isclose(getattr(self, nn), getattr(part, nn),
                    rtol=1e-14, atol=1e-14)

        for tt, nn in self._structure['per_particle_vars']:
            vv = getattr(self, nn)
            vv_copy = getattr(part, nn)[mask_copy]
            vv[i_start_copy:i_start_copy+n_copy] = vv_copy

        self.particle_id[i_start_copy:i_start_copy+n_copy] = np.arange(
                                     max_id+1, max_id+1+n_copy, dtype=np.int64)

        self.reorganize()

    def get_active_particle_id_range(self):
        ctx2np = self._buffer.context.nparray_from_context_array
        mask_active = ctx2np(self.state) > 0
        ids_active_particles = ctx2np(self.particle_id)[mask_active]
        # Behaves as python range (+1)
        return np.min(ids_active_particles), np.max(ids_active_particles)+1

    def _set_p0c(self):
        energy0 = np.sqrt(self.p0c ** 2 + self.mass0 ** 2)
        self.beta0 = self.p0c / energy0
        self.gamma0 = energy0 / self.mass0

    @property
    def delta(self):
        return self._delta

    @property
    def psigma(self):
        return self._psigma

    @property
    def rvv(self):
        return self._rvv

    @property
    def rpp(self):
        return self._rpp

    @property
    def ptau(self):
        return (
            np.sqrt(self.delta ** 2 + 2 * self.delta + 1 / self.beta0 ** 2)
            - 1 / self.beta0
        )

    @property
    def energy0(self):
        return np.sqrt( self.p0c * self.p0c + self.mass0 * self.mass0 )

    @property
    def energy(self):
        return self.energy0 + self.psigma * self.p0c * self.beta0 # eV


    def add_to_energy(self, delta_energy):
        beta0 = self.beta0.copy()
        delta_beta0 = self.delta * beta0

        ptau_beta0 = (
            delta_energy / self.energy0.copy() +
            np.sqrt( delta_beta0 * delta_beta0 + 2.0 * delta_beta0 * beta0
                    + 1. ) - 1.)

        ptau   = ptau_beta0 / beta0
        psigma = ptau / beta0
        delta = np.sqrt( ptau * ptau + 2. * psigma + 1 ) - 1

        one_plus_delta = delta + 1.
        rvv = one_plus_delta / ( 1. + ptau_beta0 )

        self.delta = delta
        self.psigma = psigma
        self.zeta *= rvv / self.rvv

        self.rvv = rvv
        self.rpp = 1. / one_plus_delta

    def set_reference(self, p0c=7e12, mass0=pmass, q0=1):
        self.q0 = q0
        self.mass0 = mass0
        self.p0c = p0c
        return self

    def set_particle(self, index, set_scalar_vars=False,
            check_scalar_vars=True, **kwargs):

        # Needed to generate consistent longitudinal variables
        pyparticles = Pyparticles(**kwargs)
        part_dict = _pyparticles_to_xpart_dict(pyparticles)
        for tt, kk in list(scalar_vars):
            setattr(self, kk, part_dict[kk])
        for tt, kk in list(per_particle_vars):
            if kk.startswith('__') and kk not in part_dict.keys():
                continue
            getattr(self, kk)[index] = part_dict[kk][0]

    def update_delta(self, new_delta_value):
        beta0 = self.beta0
        delta_beta0 = new_delta_value * beta0
        ptau_beta0  = ( delta_beta0 * delta_beta0 +
                                2. * delta_beta0 * beta0 + 1. )**0.5 - 1.
        one_plus_delta = 1. + new_delta_value
        rvv    = ( one_plus_delta ) / ( 1. + ptau_beta0 )
        rpp    = 1. / one_plus_delta
        psigma = ptau_beta0 / ( beta0 * beta0 )

        self.delta[:] = new_delta_value
        self.rvv[:] = rvv
        self.rpp[:] = rpp
        self.psigma[:] = psigma

    def compare(self, particle, rel_tol=1e-6, abs_tol=1e-15):
        identical = True
        for cc in self._structure.keys():
            for tt, nn in self._structure[cc]:
                if not np.allclose(getattr(self, nn), getattr(particle, nn),
                        atol=abs_tol, rtol=rel_tol):
                    identical = False
                    break
        return identical

def _str_in_list(string, str_list):

    found = False
    for ss in str_list:
        # TODO: Tried this but did not work
        # To avoid strange behaviors with different str formats
        # if ss.decode('utf-8') == string.decode('utf-8'):
        if ss == string:
            found = True
            break
    return found

def part_energy_varnames():
    return [vv for tt, vv in part_energy_vars]

def gen_local_particle_api(mode='no_local_copy', freeze_vars=()):

    if mode != 'no_local_copy':
        raise NotImplementedError

    src_lines = []
    src_lines.append('''typedef struct{''')
    for tt, vv in size_vars + scalar_vars:
        src_lines.append('                 ' + tt._c_type + '  '+vv+';')
    for tt, vv in per_particle_vars:
        src_lines.append('    /*gpuglmem*/ ' + tt._c_type + '* '+vv+';')
    src_lines.append(    '                 int64_t ipart;')
    src_lines.append('}LocalParticle;')
    src_typedef = '\n'.join(src_lines)

    # Particles_to_LocalParticle
    src_lines = []
    src_lines.append('''
    /*gpufun*/
    void Particles_to_LocalParticle(ParticlesData source,
                                    LocalParticle* dest,
                                    int64_t id){''')
    for tt, vv in size_vars + scalar_vars:
        src_lines.append(
                f'  dest->{vv} = ParticlesData_get_'+vv+'(source);')
    for tt, vv in per_particle_vars:
        src_lines.append(
                f'  dest->{vv} = ParticlesData_getp1_'+vv+'(source, 0);')
    src_lines.append('  dest->ipart = id;')
    src_lines.append('}')
    src_particles_to_local = '\n'.join(src_lines)

    # LocalParticle_to_Particles
    src_lines = []
    src_lines.append('''
    /*gpufun*/
    void LocalParticle_to_Particles(
                                    LocalParticle* source,
                                    ParticlesData dest,
                                    int64_t id,
                                    int64_t set_scalar){''')
    src_lines.append('if (set_scalar){')
    for tt, vv in size_vars + scalar_vars:
        src_lines.append(
                f'  ParticlesData_set_' + vv + '(dest,'
                f'      LocalParticle_get_{vv}(source));')
    src_lines.append('}')

    for tt, vv in per_particle_vars:
        src_lines.append(
                f'  ParticlesData_set_' + vv + '(dest, id, '
                f'      LocalParticle_get_{vv}(source));')
    src_lines.append('}')
    src_local_to_particles = '\n'.join(src_lines)

    # Adders
    src_lines=[]
    for tt, vv in per_particle_vars:
        src_lines.append('''
    /*gpufun*/
    void LocalParticle_add_to_'''+vv+f'(LocalParticle* part, {tt._c_type} value)'
    +'{')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('/* frozen variable!')
        src_lines.append(f'  part->{vv}[part->ipart] += value;')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('frozen variable!*/')
        src_lines.append('}\n')
    src_adders = '\n'.join(src_lines)

    # Scalers
    src_lines=[]
    for tt, vv in per_particle_vars:
        src_lines.append('''
    /*gpufun*/
    void LocalParticle_scale_'''+vv+f'(LocalParticle* part, {tt._c_type} value)'
    +'{')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('/* frozen variable!')
        src_lines.append(f'  part->{vv}[part->ipart] *= value;')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('frozen variable!*/')
        src_lines.append('}\n')
    src_scalers = '\n'.join(src_lines)

    # Setters
    src_lines=[]
    for tt, vv in per_particle_vars:
        src_lines.append('''
    /*gpufun*/
    void LocalParticle_set_'''+vv+f'(LocalParticle* part, {tt._c_type} value)'
    +'{')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('/* frozen variable!')
        src_lines.append(f'  part->{vv}[part->ipart] = value;')
        if _str_in_list(vv, freeze_vars):
            src_lines.append('frozen variable!*/')
        src_lines.append('}')
    src_setters = '\n'.join(src_lines)

    # Getters 
    src_lines=[]
    for tt, vv in size_vars + scalar_vars:
        src_lines.append('/*gpufun*/')
        src_lines.append(f'{tt._c_type} LocalParticle_get_'+vv
                        + f'(LocalParticle* part)'
                        + '{')
        src_lines.append(f'  return part->{vv};')
        src_lines.append('}')
    for tt, vv in per_particle_vars:
        src_lines.append('/*gpufun*/')
        src_lines.append(f'{tt._c_type} LocalParticle_get_'+vv
                        + f'(LocalParticle* part)'
                        + '{')
        src_lines.append(f'  return part->{vv}[part->ipart];')
        src_lines.append('}')
    src_getters = '\n'.join(src_lines)

    # Particle exchangers
    src_exchange = '''
/*gpufun*/
void LocalParticle_exchange(LocalParticle* part, int64_t i1, int64_t i2){
'''
    for tt, vv in per_particle_vars:
        src_exchange += '\n'.join([
          '\n    {',
          f'    {tt._c_type} temp = part->{vv}[i2];',
          f'    part->{vv}[i2] = part->{vv}[i1];',
          f'    part->{vv}[i1] = temp;',
          '     }'])
    src_exchange += '}\n'


    custom_source='''
/*gpufun*/
double LocalParticle_get_energy0(LocalParticle* part){

    double const p0c = LocalParticle_get_p0c(part);
    double const m0  = LocalParticle_get_mass0(part);

    return sqrt( p0c * p0c + m0 * m0 );
}

/*gpufun*/
void LocalParticle_add_to_energy(LocalParticle* part, double delta_energy){

    double const beta0 = LocalParticle_get_beta0(part);
    double const delta_beta0 = LocalParticle_get_delta(part) * beta0;

    double const ptau_beta0 =
        delta_energy / LocalParticle_get_energy0(part) +
        sqrt( delta_beta0 * delta_beta0 + 2.0 * delta_beta0 * beta0
                + 1. ) - 1.;

    double const ptau   = ptau_beta0 / beta0;
    double const psigma = ptau / beta0;
    double const delta = sqrt( ptau * ptau + 2. * psigma + 1 ) - 1;

    double const one_plus_delta = delta + 1.;
    double const rvv = one_plus_delta / ( 1. + ptau_beta0 );

    LocalParticle_set_delta(part, delta );
    LocalParticle_set_psigma(part, psigma );
    LocalParticle_scale_zeta(part,
        rvv / LocalParticle_get_rvv(part));

    LocalParticle_set_rvv(part, rvv );
    LocalParticle_set_rpp(part, 1. / one_plus_delta );
}



/*gpufun*/
void LocalParticle_update_delta(LocalParticle* part, double new_delta_value){
    double const beta0 = LocalParticle_get_beta0(part);
    double const delta_beta0 = new_delta_value * beta0;
    double const ptau_beta0  = sqrt( delta_beta0 * delta_beta0 +
                                2. * delta_beta0 * beta0 + 1. ) - 1.;

    double const one_plus_delta = 1. + new_delta_value;
    double const rvv    = ( one_plus_delta ) / ( 1. + ptau_beta0 );
    double const rpp    = 1. / one_plus_delta;
    double const psigma = ptau_beta0 / ( beta0 * beta0 );

    LocalParticle_set_delta(part, new_delta_value);

    LocalParticle_set_rvv(part, rvv );
    LocalParticle_set_rpp(part, rpp );
    LocalParticle_set_psigma(part, psigma );

}

/*gpufun*/
void LocalParticle_update_p0c(LocalParticle* part, double new_p0c_value){

    double const mass0 = LocalParticle_get_mass0(part);
    double const old_p0c = LocalParticle_get_p0c(part);
    double const old_delta = LocalParticle_get_delta(part);

    double const ppc = old_p0c * old_delta + old_p0c;
    double const new_delta = (ppc - new_p0c_value)/new_p0c_value;

    double const new_energy0 = sqrt(new_p0c_value*new_p0c_value + mass0 * mass0);
    double const new_beta0 = new_p0c_value / new_energy0;
    double const new_gamma0 = new_energy0 / mass0;

    LocalParticle_set_p0c(part, new_p0c_value);
    LocalParticle_set_gamma0(part, new_gamma0);
    LocalParticle_set_beta0(part, new_beta0);

    LocalParticle_update_delta(part, new_delta);

    LocalParticle_scale_px(part, old_p0c/new_p0c_value);
    LocalParticle_scale_py(part, old_p0c/new_p0c_value);

}
'''

    source = '\n\n'.join([src_typedef, src_adders, src_getters,
                          src_setters, src_scalers, src_exchange,
                          src_particles_to_local, src_local_to_particles,
                          custom_source])

    return source

def _pyparticles_to_xpart_dict(pyparticles):

    out = {}

    dct = pyparticles.to_dict()

    if hasattr(pyparticles, 'weight'):
        dct['weight'] = getattr(pyparticles, 'weight')
    else:
        dct['weight'] = 1.

    for tt, kk in scalar_vars + per_particle_vars:
        if kk.startswith('__'):
            continue
        # Use properties
        dct[kk] = getattr(pyparticles, kk)


    for kk, vv in dct.items():
        dct[kk] = np.atleast_1d(vv)

    lll = [len(vv) for kk, vv in dct.items() if hasattr(vv, '__len__')]
    lll = list(set(lll))
    assert len(set(lll) - {1}) <= 1
    _capacity = max(lll)
    out['_capacity'] = _capacity

    for tt, kk in scalar_vars:
        val = dct[kk]
        assert np.allclose(val, val[0], rtol=1e-10, atol=1e-14)
        out[kk] = val[0]

    for tt, kk in per_particle_vars:
        if kk.startswith('__'):
            continue

        val_py = dct[kk]

        if _capacity > 1 and len(val_py)==1:
            temp = np.zeros(int(_capacity), dtype=tt._dtype)
            temp += val_py[0]
            val_py = temp

        if type(val_py) != tt._dtype:
            val_py = np.array(val_py, dtype=tt._dtype)

        out[kk] = val_py

    #out['_num_active_particles'] = np.sum(out['state']>0)
    #out['_num_lost_particles'] = np.sum((out['state'] < 0) &
    #                                      (out['state'] > LAST_INVALID_STATE))

    return out



import logging

import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe

from .rfbucket_matching import RFBucketMatcher
from .rfbucket_matching import ThermalDistribution
from .rf_bucket import RFBucket

logger = logging.getLogger(__name__)

def _characterize_tracker(tracker, particle_ref):

    if tracker.iscollective:
        logger.warning('Ignoring collective elements in particles generation.')
        tracker = tracker._supertracker

    line = tracker.line
    T_rev = line.get_length()/(particle_ref.beta0[0]*clight)
    freq_list = []
    lag_list_deg = []
    voltage_list = []
    h_list = []
    for ee in tracker.line.elements:
        if ee.__class__.__name__ == 'Cavity':
            if ee.voltage != 0:
                freq_list.append(ee.frequency)
                lag_list_deg.append(ee.lag)
                voltage_list.append(ee.voltage)
                h_list.append(ee.frequency*T_rev)

    particle_co = tracker.find_closed_orbit(particle_ref)

    R_matrix = tracker.compute_one_turn_matrix_finite_differences(
                       particle_on_co=particle_co)

    eta = -R_matrix[4, 5]/line.get_length() # minus sign comes from z = s-ct
    alpha_mom_compaction = eta + 1/particle_ref.gamma0[0]**2

    dct={}
    dct['T_rev'] = T_rev
    dct['freq_list'] = freq_list
    dct['lag_list_deg'] = lag_list_deg
    dct['voltage_list'] = voltage_list
    dct['h_list'] = h_list
    dct['alpha_momentum_compaction'] = alpha_mom_compaction
    dct['eta'] = eta
    return dct

def generate_longitudinal_coordinates(
                                    tracker=None,
                                    particle_ref=None,
                                    mass0=None, q0=None, gamma0=None,
                                    num_particles=None,
                                    circumference=None,
                                    alpha_momentum_compaction=None,
                                    rf_harmonic=None,
                                    rf_voltage=None,
                                    rf_phase=None,
                                    p_increment=0.,
                                    distribution='gaussian',
                                    sigma_z=None
                                    ):

    if tracker is not None:
        assert particle_ref is not None
        dct = _characterize_tracker(tracker, particle_ref)

    if mass0 is None:
        assert particle_ref is not None
        mass0 = particle_ref.mass0

    if q0 is None:
        assert particle_ref is not None
        q0 = particle_ref.q0

    if gamma0 is None:
        assert particle_ref is not None
        gamma0 = particle_ref.gamma0[0]

    if circumference is None:
        assert tracker is not None
        circumference = tracker.line.get_length()

    if alpha_momentum_compaction is None:
        assert tracker is not None
        alpha_momentum_compaction = dct['alpha_momentum_compaction']

    if rf_harmonic is None:
        assert tracker is not None
        rf_harmonic=dct['h_list']

    if rf_voltage is None:
        assert tracker is not None
        rf_voltage=dct['voltage_list']

    if rf_phase is None:
        assert tracker is not None
        rf_phase=(np.array(dct['lag_list_deg']) - 180)/180*np.pi

    if distribution != 'gaussian':
        raise NotImplementedError

    assert sigma_z is not None

    rfbucket = RFBucket(circumference=circumference,
                           gamma=gamma0,
                           mass_kg=mass0/(clight**2)*qe,
                           charge_coulomb=q0*qe,
                           alpha_array=np.atleast_1d(alpha_momentum_compaction),
                           harmonic_list=np.atleast_1d(rf_harmonic),
                           voltage_list=np.atleast_1d(rf_voltage),
                           phi_offset_list=np.atleast_1d(rf_phase),
                           p_increment=p_increment)

    # Generate longitudinal coordinates
    matcher = RFBucketMatcher(rfbucket=rfbucket,
        distribution_type=ThermalDistribution,
        sigma_z=sigma_z)
    z_particles, delta_particles, _, _ = matcher.generate(
                                             macroparticlenumber=num_particles)

    return z_particles, delta_particles



import json

import numpy as np
from scipy.special import wofz
from scipy.special import gamma as tgamma

from scipy.constants import m_p as pmass_kg
from scipy.constants import e as qe
from scipy.constants import c as clight
pmass = pmass_kg*clight**2/qe

from xobjects import JEncoder

def count_not_none(*lst):
    return len(lst) - sum(p is None for p in lst)

class MathlibDefault(object):

    from numpy import sqrt, exp, sin, cos, abs, pi, tan, interp, linspace
    from numpy import power as pow

    @classmethod
    def wfun(cls, z_re, z_im):
        w = wofz(z_re + 1j * z_im)
        return w.real, w.imag

    @classmethod
    def gamma(cls, arg):
        assert arg > 0.0
        return tgamma(arg)

class Pyparticles:


    def _g1(self, mass0, p0c, energy0):
        beta0 = p0c / energy0
        gamma0 = energy0 / mass0
        return mass0, beta0, gamma0, p0c, energy0

    def _g2(self, mass0, beta0, gamma0):
        energy0 = mass0 * gamma0
        p0c = energy0 * beta0
        return mass0, beta0, gamma0, p0c, energy0

    def _f1(self, mass0, p0c):
        sqrt = self._m.sqrt
        energy0 = sqrt(p0c ** 2 + mass0 ** 2)
        return self._g1(mass0, p0c, energy0)

    def _f2(self, mass0, energy0):
        sqrt = self._m.sqrt
        p0c = sqrt(energy0 ** 2 - mass0 ** 2)
        return self._g1(mass0, p0c, energy0)

    def _f3(self, mass0, beta0):
        sqrt = self._m.sqrt
        gamma0 = 1 / sqrt(1 - beta0 ** 2)
        return self._g2(mass0, beta0, gamma0)

    def _f4(self, mass0, gamma0):
        sqrt = self._m.sqrt
        beta0 = sqrt(1 - 1 / gamma0 ** 2)
        return self._g2(mass0, beta0, gamma0)

    def copy(self, index=None):
        p = self.__class__()
        for k, v in list(self.__dict__.items()):
            if type(v) in [np.ndarray, dict]:
                if index is None:
                    v = v.copy()
                else:
                    v = v[index]
            p.__dict__[k] = v
        return p

    def __init__ref(self, p0c, energy0, gamma0, beta0):
        not_none = count_not_none(beta0, gamma0, p0c, energy0)
        if not_none == 0:
            p0c = 1e9
            not_none = 1
            # raise ValueError("Particles defined without energy reference")
        if not_none > 0:
            if p0c is not None:
                new = self._f1(self.mass0, p0c)
                self._update_ref(*new)
            elif energy0 is not None:
                new = self._f2(self.mass0, energy0)
                self._update_ref(*new)
            elif gamma0 is not None:
                new = self._f4(self.mass0, gamma0)
                self._update_ref(*new)
            elif beta0 is not None:
                new = self._f3(self.mass0, beta0)
                self._update_ref(*new)

            if not_none>1:
                ddd = {'beta0': beta0, 'gamma0': gamma0, 'energy0': energy0,
                       'p0c': p0c}
                for nn, vv in ddd.items():
                    if vv is None:
                        continue

                    if not np.allclose(vv, getattr(self, nn), atol=1e-13):
                        raise ValueError(
                            f"""\
                        Provided energy reference is inconsistent:
                        p0c    = {p0c},
                        energy0     = {energy0},
                        gamma0 = {gamma0},
                        beta0  = {beta0}"""
                        )

    def __init__delta(self, delta, ptau, psigma):
        not_none = count_not_none(delta, ptau, psigma)
        if not_none == 0:
            self.delta = 0.0
        elif not_none >= 1:
            if delta is not None:
                self.delta = delta
            elif ptau is not None:
                self.ptau = ptau
            elif psigma is not None:
                self.psigma = psigma

            if not_none>1:
                ddd = {'delta': delta, 'ptau': ptau, 'psigma': psigma}
                for nn, vv in ddd.items():
                    if vv is None:
                        continue

                    if not np.allclose(vv, getattr(self, nn), atol=1e-13):
                        raise ValueError(
                            f"""
                        Particles defined with inconsistent energy deviations:
                        delta  = {delta},
                        ptau     = {ptau},
                        psigma = {psigma}"""
                        )

    def __init__zeta(self, zeta, tau, sigma):
        not_none = count_not_none(zeta, tau, sigma)
        if not_none == 0:
            self.zeta = 0.0
        elif not_none == 1:
            if zeta is not None:
                self.zeta = zeta
            elif tau is not None:
                self.tau = tau
            elif sigma is not None:
                self.sigma = sigma
        else:
            raise ValueError(
                f"""\
            Particles defined with multiple time deviations:
            zeta  = {zeta},
            tau   = {tau},
            sigma = {sigma}"""
            )

    def __init__chi(self, mass_ratio, charge_ratio, chi):
        not_none = count_not_none(mass_ratio, charge_ratio, chi)
        if not_none == 0:
            self._chi = 1.0
            self._mass_ratio = 1.0
            self._charge_ratio = 1.0
        elif not_none == 1:
            raise ValueError(
                f"""\
            Particles defined with insufficient mass/charge information:
            chi    = {chi},
            mass_ratio = {mass_ratio},
            charge_ratio = {charge_ratio}"""
            )
        elif not_none == 2:
            if chi is None:
                self._mass_ratio = mass_ratio
                self._charge_ratio = charge_ratio
                self._chi = charge_ratio / mass_ratio
            elif mass_ratio is None:
                self._chi = chi
                self._charge_ratio = charge_ratio
                self._mass_ratio = charge_ratio / chi
            elif charge_ratio is None:
                self._chi = chi
                self._mass_ratio = mass_ratio
                self._charge_ratio = chi * mass_ratio
        else:
            self._chi = chi
            self._mass_ratio = mass_ratio
            self._charge_ratio = chi * mass_ratio
            if np.allclose(self._chi, charge_ratio / mass_ratio):
                raise ValueError(
                    f"""
            Particles defined with multiple mass/charge information:
            chi    = {chi},
            mass_ratio = {mass_ratio},
            charge_ratio = {charge_ratio}"""
                )

    def __init__(
        self,
        mathlib=None,
        **kwargs,
    ):

        for kk, vv in kwargs.items():
            if isinstance(vv, list):
                kwargs[kk] = np.array(vv)


        s = kwargs.get('s', 0.0)
        x = kwargs.get('x', 0.0)
        px = kwargs.get('px', 0.0)
        y = kwargs.get('y', 0.0)
        py = kwargs.get('py', 0.0)
        delta = kwargs.get('delta', None)
        ptau = kwargs.get('ptau', None)
        psigma = kwargs.get('psigma', None)
        rvv = kwargs.get('rvv', None)
        zeta = kwargs.get('zeta', None)
        tau = kwargs.get('tau', None)
        sigma = kwargs.get('sigma', None)
        mass0 = kwargs.get('mass0', pmass)
        q0 = kwargs.get('q0', 1.0)
        p0c = kwargs.get('p0c', None)
        energy0 = kwargs.get('energy0', None)
        gamma0 = kwargs.get('gamma0', None)
        beta0 = kwargs.get('beta0', None)
        chi = kwargs.get('chi', None)
        mass_ratio = kwargs.get('mass_ratio', None)
        charge_ratio = kwargs.get('charge_ratio', None)
        particle_id = kwargs.get('particle_id', None)
        parent_particle_id = kwargs.get('parent_particle_id', None)
        at_turn = kwargs.get('at_turn', None)
        state = kwargs.get('state', None)  # == 0 particle lost, == 1 particle active
        weight = kwargs.get('weight', None)
        at_element = kwargs.get('at_element', None)

        if mathlib is None:
            mathlib=MathlibDefault

        self._m = mathlib
        self.s = s
        self.x = x
        self.px = px
        self.y = y
        self.py = py
        self.zeta = zeta
        self._mass0 = mass0
        self.q0 = q0
        self._update_coordinates = False
        self.__init__ref(p0c, energy0, gamma0, beta0)
        self.__init__delta(delta, ptau, psigma)
        self.__init__zeta(zeta, tau, sigma)
        self.__init__chi(chi=chi, mass_ratio=mass_ratio, charge_ratio=charge_ratio)
        self._update_coordinates = True
        length = self._check_array_length()

        if particle_id is None:
            particle_id = np.arange(length) if length is not None else 0
        self.particle_id = particle_id

        if parent_particle_id is None:
            if np.isscalar(self.particle_id):
                parent_particle_id = self.particle_id
            else:
                parent_particle_id = self.particle_id.copy()
        self.parent_particle_id = parent_particle_id

        if at_turn is None:
            at_turn = np.zeros(length) if length is not None else 0
        self.at_turn = at_turn

        if at_element is None:
            at_element = np.zeros(length) if length is not None else 0
        self.at_element = at_element

        if state is None:
            state = np.ones(length) if length is not None else 1
        self.state = state

        if weight is None:
            weight = np.ones(length, dtype=np.float64) if length is not None else 1
        self.weight = weight

        self.lost_particles = []

    def _check_array_length(self):
        names = ["x", "px", "y", "py", "zeta", "delta", "_mass0", "q0", "p0c"]
        length = None
        for nn in names:
            xx = getattr(self, nn)
            if hasattr(xx, "__iter__") and len(xx)>1:
                if length is None:
                    length = len(xx)
                else:
                    if length != len(xx):
                        raise ValueError(f"invalid length len({nn})={len(xx)}")
        return length

    Px = property(lambda p: p.px * p.p0c * p.mass_ratio)
    Py = property(lambda p: p.py * p.p0c * p.mass_ratio)
    energy = property(lambda p: (p.ptau * p.p0c + p.energy0) * p.mass_ratio)
    pc = property(lambda p: (p.delta * p.p0c + p.p0c) * p.mass_ratio)
    mass = property(lambda p: p.mass0 * p.mass_ratio)
    beta = property(lambda p: (1 + p.delta) / (1 / p.beta0 + p.ptau))
    # rvv = property(lambda self: self.beta/self.beta0)
    # rpp = property(lambda self: 1/(1+self.delta))

    rvv = property(lambda self: self._rvv)
    rpp = property(lambda self: self._rpp)

    def add_to_energy(self, energy):
        sqrt = self._m.sqrt
        oldrvv = self._rvv
        deltabeta0 = self.delta * self.beta0
        ptaubeta0 = sqrt(deltabeta0 ** 2 + 2 * deltabeta0 * self.beta0 + 1) - 1
        ptaubeta0 += energy / self.energy0
        ptau = ptaubeta0 / self.beta0
        self._delta = sqrt(ptau ** 2 + 2 * ptau / self.beta0 + 1) - 1
        self._rvv = (1 + self.delta) / (1 + ptaubeta0)
        self._rpp = 1 / (1 + self.delta)
        self.zeta *= self._rvv / oldrvv

    delta = property(lambda self: self._delta)

    @delta.setter
    def delta(self, delta):
        sqrt = self._m.sqrt
        if hasattr(self, '_rvv'):
            oldrvv = self._rvv
        else:
            # Needed for particle initialization
            oldrvv = None
        self._delta = delta
        deltabeta0 = delta * self.beta0
        ptaubeta0 = sqrt(deltabeta0 ** 2 + 2 * deltabeta0 * self.beta0 + 1) - 1
        self._rvv = (1 + self.delta) / (1 + ptaubeta0)
        self._rpp = 1 / (1 + self.delta)
        if oldrvv is not None:
            self.zeta *= self._rvv / oldrvv

    psigma = property(lambda self: self.ptau / self.beta0)

    @psigma.setter
    def psigma(self, psigma):
        self.ptau = psigma * self.beta0

    tau = property(lambda self: self.zeta / self.beta)

    @tau.setter
    def tau(self, tau):
        self.zeta = self.beta * tau

    sigma = property(lambda self: (self.beta0 / self.beta) * self.zeta)

    @sigma.setter
    def sigma(self, sigma):
        self.zeta = self.beta / self.beta0 * sigma

    @property
    def ptau(self):
        sqrt = self._m.sqrt
        return (
            sqrt(self.delta ** 2 + 2 * self.delta + 1 / self.beta0 ** 2)
            - 1 / self.beta0
        )

    @ptau.setter
    def ptau(self, ptau):
        sqrt = self._m.sqrt
        self.delta = sqrt(ptau ** 2 + 2 * ptau / self.beta0 + 1) - 1

    mass0 = property(lambda self: self._mass0)

    @mass0.setter
    def mass0(self, mass0):
        new = self._f1(mass0, self.p0c)
        _abs = self._get_absolute()
        self._update_ref(*new)
        self._update_particles_from_absolute(*_abs)

    beta0 = property(lambda self: self._beta0)

    @beta0.setter
    def beta0(self, beta0):
        new = self._f3(self.mass0, beta0)
        _abs = self._get_absolute()
        self._update_ref(*new)
        self._update_particles_from_absolute(*_abs)

    gamma0 = property(lambda self: self._gamma0)

    @gamma0.setter
    def gamma0(self, gamma0):
        new = self._f4(self.mass0, gamma0)
        _abs = self._get_absolute()
        self._update_ref(*new)
        self._update_particles_from_absolute(*_abs)

    p0c = property(lambda self: self._p0c)

    @p0c.setter
    def p0c(self, p0c):
        new = self._f1(self.mass0, p0c)
        _abs = self._get_absolute()
        self._update_ref(*new)
        self._update_particles_from_absolute(*_abs)

    energy0 = property(lambda self: self._energy0)

    @energy0.setter
    def energy0(self, energy0):
        new = self._f2(self.mass0, energy0)
        _abs = self._get_absolute()
        self._update_ref(*new)
        self._update_particles_from_absolute(*_abs)

    mass_ratio = property(lambda self: self._mass_ratio)

    @mass_ratio.setter
    def mass_ratio(self, mass_ratio):
        self._mass_ratio = mass_ratio
        self._chi = self._charge_ratio / self._mass_ratio

    charge_ratio = property(lambda self: self._charge_ratio)

    @charge_ratio.setter
    def charge_ratio(self, charge_ratio):
        self._charge_ratio = charge_ratio
        self._chi = charge_ratio / self._mass_ratio

    chi = property(lambda self: self._chi)

    @chi.setter
    def chi(self, chi):
        self._charge_ratio = self._chi * self._mass_ratio
        self._chi = chi

    def _get_absolute(self):
        return self.Px, self.Py, self.pc, self.energy

    def _update_ref(self, mass0, beta0, gamma0, p0c, energy0):
        self._mass0 = mass0
        self._beta0 = beta0
        self._gamma0 = gamma0
        self._p0c = p0c
        self._energy0 = energy0

    def _update_particles_from_absolute(self, Px, Py, pc, energy):
        if self._update_coordinates:
            mass_ratio = self.mass / self.mass0
            norm = mass_ratio * self.p0c
            self._mass_ratio = mass_ratio
            self._chi = self._charge_ratio / mass_ratio
            self._ptau = energy / norm - 1
            self._delta = pc / norm - 1
            self.px = Px / norm
            self.py = Py / norm

    def __repr__(self):
        out = f"""\
        mass0   = {self.mass0}
        p0c     = {self.p0c}
        energy0 = {self.energy0}
        beta0   = {self.beta0}
        gamma0  = {self.gamma0}
        s       = {self.s}
        x       = {self.x}
        px      = {self.px}
        y       = {self.y}
        py      = {self.py}
        zeta    = {self.zeta}
        delta   = {self.delta}
        ptau    = {self.ptau}
        mass_ratio  = {self.mass_ratio}
        charge_ratio  = {self.charge_ratio}
        chi     = {self.chi}
        state   = {self.state}
        weight  = {self.weight}"""
        return out

    _dict_vars = (
        "s",
        "x",
        "px",
        "y",
        "py",
        "zeta",
        "delta",
        "mass0",
        "q0",
        "p0c",
        "chi",
        "mass_ratio",
        "particle_id",
        "parent_particle_id",
        "at_turn",
        "state",
        "weight",
    )

    def remove_lost_particles(self, keep_memory=True):

        if hasattr(self.state, "__iter__"):
            mask_valid = self.state == 1

            if np.any(~mask_valid):
                if keep_memory:
                    to_trash = self.copy()  # Not exactly efficient (but robust)
                    for ff in ('_rvv',) + self._dict_vars: # _rvv needs to be updated
                                                           # before delta
                        if hasattr(getattr(self, ff), "__iter__"):
                            setattr(to_trash, ff, getattr(self, ff)[~mask_valid])
                    self.lost_particles.append(to_trash)

            for ff in ('_rvv',) + self._dict_vars:
                if hasattr(getattr(self, ff), "__iter__"):
                    setattr(self, ff, getattr(self, ff)[mask_valid])

    def to_dict(self):
        return {kk: getattr(self, kk) for kk in self._dict_vars}

    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)

    def to_json(self, filename):
        with open(filename, 'w') as fid:
            json.dump(self.to_dict(), fid, cls=JEncoder)

    @classmethod
    def from_json(cls, filename):
        with open(filename, 'r') as fid:
            return cls.from_dict(json.load(fid))

    def compare(self, particle, rel_tol=1e-6, abs_tol=1e-15):
        res = True
        for kk in self._dict_vars:
            v1 = getattr(self, kk)
            v2 = getattr(particle, kk)
            if v1 is not None and v2 is not None:
                diff = v1 - v2
                if hasattr(diff, "__iter__"):
                    for nn in range(len(diff)):
                        vv1 = v1[nn] if hasattr(v1, "__iter__") else v1
                        vv2 = v2[nn] if hasattr(v2, "__iter__") else v2
                        if abs(diff[nn]) > abs_tol:
                            print(f"{kk}[{nn}] {vv1} {vv2}  diff:{diff[nn]}")
                            res = False
                        if abs(vv1) > 0 and abs(diff[nn]) / vv1 > rel_tol:
                            print(f"{kk}[{nn}] {vv1} {vv2} rdiff:{diff[nn]/vv1}")
                            res = False
                else:
                    if abs(diff) > abs_tol:
                        print(f"{kk} {v1} {v2}  diff:{diff}")
                        res = False
                    if abs(v1) > 0 and abs(diff) / v1 > rel_tol:
                        print(f"{kk} {v1} {v2} rdiff:{diff/v1}")
                        res = False
        return res

    @classmethod
    def from_madx_twiss(cls, twiss):
        out = cls(
            p0c=twiss.summary.pc * 1e6,
            mass0=twiss.summary.mass * 1e6,
            q0=twiss.summary.charge,
            s=twiss.s[:],
            x=twiss.x[:],
            px=twiss.px[:],
            y=twiss.y[:],
            py=twiss.py[:],
            tau=twiss.t[:],
            ptau=twiss.pt[:],
        )
        return out

    @classmethod
    def from_madx_track(cls, mad):
        tracksumm = mad.table.tracksumm
        mad_beam = mad.sequence().beam
        out = cls(
            p0c=mad_beam.pc * 1e6,
            mass0=mad_beam.mass * 1e6,
            q0=mad_beam.charge,
            s=tracksumm.s[:],
            x=tracksumm.x[:],
            px=tracksumm.px[:],
            y=tracksumm.y[:],
            py=tracksumm.py[:],
            tau=tracksumm.t[:],
            ptau=tracksumm.pt[:],
        )
        return out

    @classmethod
    def from_list(cls, lst):
        ll = len(lst)
        dct = {nn: np.zeros(ll) for nn in cls._dict_vars}
        for ii, pp in enumerate(lst):
            for nn in cls._dict_vars:
                dct[nn][ii] = getattr(pp, nn, 0)
        return cls(**dct)



import itertools

import numpy as np

def _configure_grid(vname, v_grid, dv, v_range, nv):

    # Check input consistency
    if v_grid is not None:
        assert dv is None, (f'd{vname} cannot be given '
                            f'if {vname}_grid is provided ')
        assert nv is None, (f'n{vname} cannot be given '
                            f'if {vname}_grid is provided ')
        assert v_range is None, (f'{vname}_range cannot be given '
                                 f'if {vname}_grid is provided')
        ddd = np.diff(v_grid)
        assert np.allclose(ddd,ddd[0]), (f'{vname}_grid must be '
                                          'unifirmly spaced')
    else:
        assert v_range is not None, (f'{vname}_grid or {vname}_range '
                                     f'must be provided')
        assert len(v_range)==2, (f'{vname}_range must be in the form '
                                 f'({vname}_min, {vname}_max)')
        if dv is not None:
            assert nv is None, (f'n{vname} cannot be given '
                                    f'if d{vname} is provided ')
            v_grid = np.arange(v_range[0], v_range[1]+0.1*dv, dv)
        else:
            assert nv is not None, (f'n{vname} must be given '
                                    f'if d{vname} is not provided ')
            v_grid = np.linspace(v_range[0], v_range[1], nv)

    return v_grid


def generate_2D_polar_grid(
        r_range=None, r_grid=None, dr=None, nr=None,
        theta_range=None, theta_grid=None, dtheta=None, ntheta=None):

    _r_grid = _configure_grid('r', r_grid, dr, r_range, nr)
    _theta_grid = _configure_grid('theta', theta_grid, dtheta,
                                  theta_range, ntheta)

    temp = np.array([(ii[0], ii[1]) for ii in
                    itertools.product(_r_grid, _theta_grid)])

    r_all = temp[:, 0]
    theta_all = temp[:, 1]

    a1 = r_all*np.cos(theta_all)
    a2 = r_all*np.sin(theta_all)

    return a1, a2, r_all, theta_all

def generate_2D_uniform_circular_sector(num_particles, r_range=(0, 1),
                                        theta_range=(0, 2*np.pi)):


    # CDF(r) = (r^2 - r0^2)/(r1^2 - r0^2)
    # InvCDF(u) = sqrt(r0^2 + u * (r1^2 -r0^2))

    r0 = r_range[0]
    r1 = r_range[1]

    uu = np.random.uniform(low=0, high=1., size=num_particles)

    r_all = np.sqrt(r0*r0 + uu * (r1*r1 - r0*r0))

    theta_all = np.random.uniform(low=theta_range[0], high=theta_range[1],
                           size=num_particles)

    a1 = r_all*np.cos(theta_all)
    a2 = r_all*np.sin(theta_all)

    return a1, a2, r_all, theta_all



import numpy as np

from .longitudinal import generate_longitudinal_coordinates
from .linear_normal_form import compute_linear_normal_form
from .build_particles import build_particles
from .particles import Particles


def generate_matched_gaussian_bunch(num_particles, total_intensity_particles,
                                    nemitt_x, nemitt_y, sigma_z,
                                    particle_on_co=None,
                                    R_matrix=None,
                                    circumference=None,
                                    alpha_momentum_compaction=None,
                                    rf_harmonic=None,
                                    rf_voltage=None,
                                    rf_phase=None,
                                    p_increment=0.,
                                    tracker=None,
                                    particle_ref=None,
                                    particle_class=Particles,
                                    _context=None, _buffer=None, _offset=None,
                                    ):

    if (particle_ref is not None and particle_on_co is not None):
        raise ValueError("`particle_ref` and `particle_on_co`"
                " cannot be provided at the same time")

    if particle_ref is None:
        assert particle_on_co is not None, (
            "`particle_ref` or `particle_on_co` must be provided!")
        particle_ref = particle_on_co

    zeta, delta = generate_longitudinal_coordinates(
            distribution='gaussian',
            num_particles=num_particles,
            particle_ref=(particle_ref if particle_ref is not None
                          else particle_on_co),
            tracker=tracker,
            circumference=circumference,
            alpha_momentum_compaction=alpha_momentum_compaction,
            rf_harmonic=rf_harmonic,
            rf_voltage=rf_voltage,
            rf_phase=rf_phase,
            p_increment=p_increment,
            sigma_z=sigma_z)

    assert len(zeta) == len(delta) == num_particles

    x_norm = np.random.normal(size=num_particles)
    px_norm = np.random.normal(size=num_particles)
    y_norm = np.random.normal(size=num_particles)
    py_norm = np.random.normal(size=num_particles)


    part = build_particles(_context=_context, _buffer=_buffer, _offset=_offset,
                      R_matrix=R_matrix,
                      particle_class=particle_class,
                      particle_on_co=particle_on_co,
                      particle_ref=(
                          particle_ref if particle_on_co is  None else None),
                      tracker=tracker,
                      zeta=zeta, delta=delta,
                      x_norm=x_norm, px_norm=px_norm,
                      y_norm=y_norm, py_norm=py_norm,
                      scale_with_transverse_norm_emitt=(nemitt_x, nemitt_y),
                      weight=total_intensity_particles/num_particles)
    return part



from .general import _pkg_root
from .particles import Particles, gen_local_particle_api, pmass

from .build_particles import build_particles
from .matched_gaussian import generate_matched_gaussian_bunch

from .linear_normal_form import compute_linear_normal_form

from .transverse_generators import generate_2D_polar_grid
from .transverse_generators import generate_2D_uniform_circular_sector
from .transverse_generators import generate_2D_pencil
from .transverse_generators import generate_2D_gaussian

from .longitudinal import generate_longitudinal_coordinates

from .constants import PROTON_MASS_EV, ELECTRON_MASS_EV

def enable_pyheadtail_interface():
    import xpart.pyheadtail_interface.pyhtxtparticles as pp
    import xpart as xp
    xp.Particles = pp.PyHtXtParticles



import numpy as np
from scipy.constants import e,c

from PyHEADTAIL.particles.particles import Particles as PyHtParticles
from xpart import Particles as XtParticles


class PyHtXtParticles(XtParticles,PyHtParticles):

    def __init__(self,circumference=None,particlenumber_per_mp=1.0, **kwargs):
        super(PyHtXtParticles,self).__init__(**kwargs)
        self.circumference = circumference
        self.particlenumber_per_mp = particlenumber_per_mp
        self._slice_sets = {}
        self.coords_n_momenta = set(['x','xp','y','yp','z','dp'])

    @classmethod
    def from_pyheadtail(cls, particles):
        new  = cls(_capacity=particles.macroparticlenumber)

        new.particlenumber_per_mp = particles.particlenumber_per_mp
        new.charge = particles.charge
        new.mass = particles.mass
        new.circumference = particles.circumference
        new.gamma = particles.gamma
        new.x = particles.x
        new.xp = particles.xp
        new.y = particles.y
        new.yp = particles.yp
        new.z = particles.z
        new.dp = particles.dp

        return new

    @property
    def z(self):
        return self.zeta

    @z.setter
    def z(self, value):
        self.zeta = value

    @property
    def xp(self):
        return self.px

    @xp.setter
    def xp(self, value):
        self.px = value

    @property
    def yp(self):
        return self.py

    @yp.setter
    def yp(self, value):
        self.py = value

    @property
    def dp(self):
        return self.delta

    @dp.setter
    def dp(self, value):
        self.update_delta(value)

    @property
    def mass(self):
        return self.mass0/(c*c)*e

    @mass.setter
    def mass(self, value):
        self.mass0 = value/e*c*c

    @property
    def charge(self):
        return self.q0*e

    @charge.setter
    def charge(self, value):
        self.q0 = value/e

    @property
    def macroparticlenumber(self):
        return np.sum(self.state>0)

    @property
    def particlenumber_per_mp(self):
        return self.weight[0] # I avoid checking that they are all the same
                         # not to compromise on performance

    @particlenumber_per_mp.setter
    def particlenumber_per_mp(self, value):
        self.weight[:] = value

    @property
    def _gamma(self):
	# I assume that they are all the same and take the first
        # An assert would be too expensive...
        return self.gamma0[0]

    @_gamma.setter
    def _gamma(self, value):
        self.gamma0[:] = value

    @property
    def _beta(self):
	# I assume that they are all the same and take the first
        # An assert would be too expensive...
        return self.beta0[0]

    @_beta.setter
    def _beta(self, value):
        self.beta0[:] = value

    @property
    def _p0(self):
	# I assume that they are all the same and take the first
        # An assert would be too expensive...
        return self.p0c[0]/c*e

    @_p0.setter
    def _p0(self, value):
        self.p0c[:] = value/e*c

    @property
    def id(self):
        return self.particle_id




import logging

import numpy as np

import xobjects as xo

from .linear_normal_form import compute_linear_normal_form
from .particles import Particles

logger = logging.getLogger(__name__)

def _check_lengths(**kwargs):
    length = None
    for nn, xx in kwargs.items():
        if hasattr(xx, "__iter__"):
            if length is None:
                length = len(xx)
            else:
                if length != len(xx):
                    raise ValueError(f"invalid length len({nn})={len(xx)}")
    if 'num_particles' in kwargs.keys():
        num_particles = kwargs['num_particles']
        if num_particles is not None and length is None:
            length = num_particles
        if num_particles is not None and length != num_particles:
            raise ValueError(
              f"num_particles={num_particles} is inconsistent with array length")

    if length is None:
        length = 1
    return length

def build_particles(_context=None, _buffer=None, _offset=None, _capacity=None,
                      mode=None,
                      particle_ref=None,
                      num_particles=None,
                      x=None, px=None, y=None, py=None, zeta=None, delta=None,
                      x_norm=None, px_norm=None, y_norm=None, py_norm=None,
                      tracker=None,
                      particle_on_co=None,
                      R_matrix=None,
                      scale_with_transverse_norm_emitt=None,
                      weight=None,
                      particle_class=Particles,
                    ):

    """
    Function to create particle objects from arrays containing physical or
    normalized coordinates.

    Arguments:

        - mode: choose between:

            - `set`: reference quantities including mass0, q0, p0c, gamma0,
              etc. are taken from the provided reference particle. Particles
              coordinates are set according to the provided input x, px, y, py,
              zeta, delta (zero is assumed as default for these variables).
            - `shift`: reference quantities including mass0, q0, p0c, gamma0,
              etc. are taken from the provided reference particle. Particles
              coordinates are set from the reference particles and shifted
              according to the provided input x, px, y, py, zeta, delta (zero
              is assumed as default for these variables).
            - `normalized_transverse`: reference quantities including mass0,
              q0, p0c, gamma0, etc. are taken from the provided reference
              particle. The longitudinal coordinates are set according to the
              provided input `zeta`, `delta` (zero is assumed as default value
              for these variable`. The transverse coordinates are computed from
              normalized values `x_norm`, `px_norm`, `y_norm`, `py_norm` using
              the closed-orbit information and the linear transfer map obtained
              from the `tracker` or provided by the user.

            The default mode is `set`. `normalized_transverse` is used if any
            of x_norm, px_norm, y_norm, pynorm is provided.
        - particle_ref: particle object defining the reference quantities
          (mass0, 0, p0c, gamma0, etc.). Its coordinates (x, py, y, py, zeta,
          delta) are ignored unless `mode`='shift' is selected.
        - num_particles: Number of particles to be generated (used if provided
          coordinates are all scalar)
        - x: x coordinate of the particles (default is 0).
        - px: px coordinate of the particles (default is 0).
        - y: y coordinate of the particles (default is 0).
        - py: py coordinate of the particles (default is 0).
        - zeta: zeta coordinate of the particles (default is 0).
        - delta: delta coordinate of the particles (default is 0).
        - x_norm: transverse normalized coordinate x (in sigmas) used in
            combination with the one turn matrix R_matrix and with the
            transverse emittances provided in the argument
            `scale_with_transverse_norm_emitt` to generate x, px, y, py (x, px,
            y, py cannot be provided if x_norm, px_norm, y_norm, py_norm are
            provided).
        - x_norm: transverse normalized coordinate x (in sigmas).
        - px_norm: transverse normalized coordinate px (in sigmas).
        - y_norm: transverse normalized coordinate y (in sigmas).
        - py_norm: transverse normalized coordinate py (in sigmas).
        - tracker: tracker object used to find the closed orbit and the
          one-turn matrix.
        - particle_on_co: Particle on closed orbit
        - R_matrix: 6x6 matrix defining the linearized one-turn map to be used
          for the transformation of the normalized coordinates into physical
          space.
        - scale_with_transverse_norm_emitt: Tuple of two elements defining the
          transverse normalized emittances used to rescale the provided
          transverse normalized coordinates (x, px, y, py).
        - weight: weights to be assigned to the particles.
        - _context: xobjects context in which the particle object is allocated.

    """

    assert mode in [None, 'set', 'shift', 'normalized_transverse']

    if (particle_ref is not None and particle_on_co is not None):
        raise ValueError("`particle_ref` and `particle_on_co`"
                " cannot be provided at the same time")

    if particle_ref is None:
        assert particle_on_co is not None, (
            "`particle_ref` or `particle_on_co` must be provided!")
        particle_ref = particle_on_co

    if not isinstance(particle_ref._buffer.context, xo.ContextCpu):
        particle_ref = particle_ref.copy(_context=xo.ContextCpu())

    if tracker is not None and tracker.iscollective:
        logger.warning('Ignoring collective elements in particles generation.')
        tracker = tracker._supertracker

    if zeta is None:
        zeta = 0

    if delta is None:
        delta = 0

    if (x_norm is not None or px_norm is not None
        or y_norm is not None or py_norm is not None):

        assert (x is  None and px is  None
                and y is  None and py is  None)

        if mode is None:
            mode = 'normalized_transverse'
        else:
            assert mode == 'normalized_transverse'

    if mode is None:
        mode = 'set'

    if mode == 'normalized_transverse':
        if x_norm is None: x_norm = 0
        if px_norm is None: px_norm = 0
        if y_norm is None: y_norm = 0
        if py_norm is None: py_norm = 0
    else:
        if x is None: x = 0
        if px is None: px = 0
        if y is None: y = 0
        if py is None: py = 0

    assert particle_ref._capacity == 1
    ref_dict = {
        'q0': particle_ref.q0,
        'mass0': particle_ref.mass0,
        'p0c': particle_ref.p0c[0],
        'gamma0': particle_ref.gamma0[0],
        'beta0': particle_ref.beta0[0],
    }
    part_dict = ref_dict.copy()

    if mode == 'normalized_transverse':
        if particle_on_co is None:
            assert tracker is not None
            particle_on_co = tracker.find_closed_orbit(
                particle_co_guess=Particles(
                    x=0, px=0, y=0, py=0, zeta=0, delta=0.,
                    **ref_dict))
        else:
            assert particle_on_co._capacity == 1

        if not isinstance(particle_on_co._buffer.context, xo.ContextCpu):
            particle_on_co = particle_on_co.copy(_context=xo.ContextCpu())

        if R_matrix is None:
            R_matrix = tracker.compute_one_turn_matrix_finite_differences(
                particle_on_co=particle_on_co)

        num_particles = _check_lengths(num_particles=num_particles,
            zeta=zeta, delta=delta, x_norm=x_norm, px_norm=px_norm,
            y_norm=y_norm, py_norm=py_norm)

        if scale_with_transverse_norm_emitt is not None:
            assert len(scale_with_transverse_norm_emitt) == 2

            nemitt_x = scale_with_transverse_norm_emitt[0]
            nemitt_y = scale_with_transverse_norm_emitt[1]

            gemitt_x = nemitt_x/particle_ref.beta0/particle_ref.gamma0
            gemitt_y = nemitt_y/particle_ref.beta0/particle_ref.gamma0

            x_norm_scaled = np.sqrt(gemitt_x) * x_norm
            px_norm_scaled = np.sqrt(gemitt_x) * px_norm
            y_norm_scaled = np.sqrt(gemitt_y) * y_norm
            py_norm_scaled = np.sqrt(gemitt_y) * py_norm
        else:
            x_norm_scaled = x_norm
            px_norm_scaled = px_norm
            y_norm_scaled = y_norm
            py_norm_scaled = py_norm

        WW, WWinv, Rot = compute_linear_normal_form(R_matrix)

        # Transform long. coordinates to normalized space
        XX_long = np.zeros(shape=(6, num_particles), dtype=np.float64)
        XX_long[4, :] = zeta - particle_on_co.zeta
        XX_long[5, :] = delta - particle_on_co.delta

        XX_norm_scaled = np.dot(WWinv, XX_long)

        XX_norm_scaled[0, :] = x_norm_scaled
        XX_norm_scaled[1, :] = px_norm_scaled
        XX_norm_scaled[2, :] = y_norm_scaled
        XX_norm_scaled[3, :] = py_norm_scaled

        # Transform to physical coordinates
        XX = np.dot(WW, XX_norm_scaled)

        XX[0, :] += particle_on_co.x
        XX[1, :] += particle_on_co.px
        XX[2, :] += particle_on_co.y
        XX[3, :] += particle_on_co.py
        XX[4, :] += particle_on_co.zeta
        XX[5, :] += particle_on_co.delta

    elif mode == 'set':

        if R_matrix is not None:
            logger.warning('R_matrix provided but not used in this mode!')

        num_particles = _check_lengths(num_particles=num_particles,
            zeta=zeta, delta=delta, x=x, px=px,
            y=y, py=py)

        XX = np.zeros(shape=(6, num_particles), dtype=np.float64)
        XX[0, :] = x
        XX[1, :] = px
        XX[2, :] = y
        XX[3, :] = py
        XX[4, :] = zeta
        XX[5, :] = delta

    elif mode == "shift":

        if R_matrix is not None:
            logger.warning('R_matrix provided but not used in this mode!')

        num_particles = _check_lengths(num_particles=num_particles,
            zeta=zeta, delta=delta, x=x, px=px,
            y=y, py=py)

        XX = np.zeros(shape=(6, num_particles), dtype=np.float64)
        XX[0, :] = x + particle_ref.x
        XX[1, :] = px + particle_ref.px
        XX[2, :] = y + particle_ref.y
        XX[3, :] = py + particle_ref.py
        XX[4, :] = zeta + particle_ref.zeta
        XX[5, :] = delta + particle_ref.delta
    else:
        raise ValueError('What?!')

    part_dict['x'] = XX[0, :]
    part_dict['px'] = XX[1, :]
    part_dict['y'] = XX[2, :]
    part_dict['py'] = XX[3, :]
    part_dict['zeta'] = XX[4, :]
    part_dict['delta'] = XX[5, :]

    part_dict['weight'] = np.zeros(num_particles, dtype=np.int64)

    particles = Particles(_context=_context, _buffer=_buffer, _offset=_offset,
                          _capacity=_capacity,**part_dict)

    particles.particle_id = particles._buffer.context.nparray_to_context_array(
                                   np.arange(0, num_particles, dtype=np.int64))
    if weight is not None:
        particles.weight[:] = weight

    return particles



import numpy as np

from .particles import Particles


def healy_symplectify(M):
    # https://accelconf.web.cern.ch/e06/PAPERS/WEPCH152.PDF
    print("Symplectifying linear One-Turn-Map...")

    print("Before symplectifying: det(M) = {}".format(np.linalg.det(M)))
    I = np.identity(6)

    S = np.array(
        [
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, -1.0, 0.0],
        ]
    )

    V = np.matmul(S, np.matmul(I - M, np.linalg.inv(I + M)))
    W = (V + V.T) / 2
    if np.linalg.det(I - np.matmul(S, W)) != 0:
        M_new = np.matmul(I + np.matmul(S, W),
                          np.linalg.inv(I - np.matmul(S, W)))
    else:
        print("WARNING: det(I - SW) = 0!")
        V_else = np.matmul(S, np.matmul(I + M, np.linalg.inv(I - M)))
        W_else = (V_else + V_else.T) / 2
        M_new = -np.matmul(
            I + np.matmul(S, W_else), np.linalg(I - np.matmul(S, W_else))
        )

    print("After symplectifying: det(M) = {}".format(np.linalg.det(M_new)))
    return M_new

S = np.array([[0., 1., 0., 0., 0., 0.],
              [-1., 0., 0., 0., 0., 0.],
              [ 0., 0., 0., 1., 0., 0.],
              [ 0., 0.,-1., 0., 0., 0.],
              [ 0., 0., 0., 0., 0., 1.],
              [ 0., 0., 0., 0.,-1., 0.]])

######################################################
### Implement Normalization of fully coupled motion ##
######################################################

def Rot2D(mu):
    return np.array([[ np.cos(mu), np.sin(mu)],
                     [-np.sin(mu), np.cos(mu)]])

def compute_linear_normal_form(M, symplectify=True, tol_det_M=1e-3):

    if np.abs(np.linalg.det(M)-1) > tol_det_M:
        raise ValueError('The determinant of M is out tolerance.')

    if symplectify:
        M = healy_symplectify(M)

    w0, v0 = np.linalg.eig(M)

    a0 = np.real(v0)
    b0 = np.imag(v0)

    index_list = [0,5,1,2,3,4]

    ##### Sort modes in pairs of conjugate modes #####

    conj_modes = np.zeros([3,2], dtype=np.int64)
    for j in [0,1]:
        conj_modes[j,0] = index_list[0]
        del index_list[0]

        min_index = 0
        min_diff = abs(np.imag(w0[conj_modes[j,0]] + w0[index_list[min_index]]))
        for i in range(1,len(index_list)):
            diff = abs(np.imag(w0[conj_modes[j,0]] + w0[index_list[i]]))
            if min_diff > diff:
                min_diff = diff
                min_index = i

        conj_modes[j,1] = index_list[min_index]
        del index_list[min_index]

    conj_modes[2,0] = index_list[0]
    conj_modes[2,1] = index_list[1]

    ##################################################
    #### Select mode from pairs with positive (real @ S @ imag) #####

    modes = np.empty(3, dtype=np.int64)
    for ii,ind in enumerate(conj_modes):
        if np.matmul(np.matmul(a0[:,ind[0]], S), b0[:,ind[0]]) > 0:
            modes[ii] = ind[0]
        else:
            modes[ii] = ind[1]

    ##################################################
    #### Sort modes such that (1,2,3) is close to (x,y,zeta) ####
    for i in [1,2]:
        if abs(v0[:,modes[0]])[0] < abs(v0[:,modes[i]])[0]:
            modes[0], modes[i] = modes[i], modes[0]

    if abs(v0[:,modes[1]])[2] < abs(v0[:,modes[2]])[2]:
        modes[2], modes[1] = modes[1], modes[2]

    ##################################################
    #### Rotate eigenvectors to the Courant-Snyder parameterization ####
    phase0 = np.log(v0[0,modes[0]]).imag
    phase1 = np.log(v0[2,modes[1]]).imag
    phase2 = np.log(v0[4,modes[2]]).imag

    v0[:,modes[0]] *= np.exp(-1.j*phase0)
    v0[:,modes[1]] *= np.exp(-1.j*phase1)
    v0[:,modes[2]] *= np.exp(-1.j*phase2)

    ##################################################
    #### Construct W #################################

    a1 = v0[:,modes[0]].real
    a2 = v0[:,modes[1]].real
    a3 = v0[:,modes[2]].real
    b1 = v0[:,modes[0]].imag
    b2 = v0[:,modes[1]].imag
    b3 = v0[:,modes[2]].imag

    n1 = 1./np.sqrt(np.matmul(np.matmul(a1, S), b1))
    n2 = 1./np.sqrt(np.matmul(np.matmul(a2, S), b2))
    n3 = 1./np.sqrt(np.matmul(np.matmul(a3, S), b3))

    a1 *= n1
    a2 *= n2
    a3 *= n3

    b1 *= n1
    b2 *= n2
    b3 *= n3

    W = np.array([a1,b1,a2,b2,a3,b3]).T
    W[abs(W) < 1.e-14] = 0. # Set very small numbers to zero.
    invW = np.matmul(np.matmul(S.T, W.T), S)

    ##################################################
    #### Get tunes and rotation matrix in the normalized coordinates ####

    mu1 = np.log(w0[modes[0]]).imag
    mu2 = np.log(w0[modes[1]]).imag
    mu3 = np.log(w0[modes[2]]).imag

    #q1 = mu1/(2.*np.pi)
    #q2 = mu2/(2.*np.pi)
    #q3 = mu3/(2.*np.pi)

    R = np.zeros_like(W)
    R[0:2,0:2] = Rot2D(mu1)
    R[2:4,2:4] = Rot2D(mu2)
    R[4:6,4:6] = Rot2D(mu3)
    ##################################################

    return W, invW, R



from setuptools import setup, find_packages, Extension

#######################################
# Prepare list of compiled extensions #
#######################################

extensions = []


#########
# Setup #
#########

setup(
    name='xpart',
    version='0.3.2',
    description='Generation of Particle Ensembles',
    url='https://github.com/xsuite/xpart',
    packages=find_packages(),
    ext_modules = extensions,
    include_package_data=True,
    install_requires=[
        'numpy>=1.0',
        'scipy',
        'xobjects'
        ]
    )



'''
@author Kevin Li, Adrian Oeftiger
@date 21.06.2017
@brief 2D distribution integration methods for y(x) parametrised domains
'''

import numpy as np
from scipy.integrate import quad, dblquad, cumtrapz, romb


def quad2d(f, ylimits, xmin, xmax):
    '''Integrate 2D function f=f(x,y) over interval [xmin,xmax] and
    between contours [-ylimits(x), ylimits(x)] using the numerical
    scipy.integrate.dblquad integration method.
    '''
    Q, error = dblquad(lambda y, x: f(x, y), xmin, xmax,
                       lambda x: -ylimits(x), lambda x: ylimits(x))

    return Q

### scipy.integrate.dblquad moment integration:

def compute_zero_quad(psi, ylimit_min, ylimit_max, xmin, xmax):
    '''Compute the zeroth moment of the distribution function psi from
    xmin to xmax between the contours ylimit_min and ylimit_max
    (e.g. an RFBucket separatrix) using the numerical
    scipy.integrate.dblquad integration method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
    '''
    Q, error = dblquad(lambda y, x: psi(x, y), xmin, xmax,
                       ylimit_min, ylimit_max)

    return Q

def compute_mean_quad(psi, ylimit_min, ylimit_max, xmin, xmax, direction='x'):
    '''Compute the first moment of the distribution function psi from
    xmin to xmax between the contours ylimit_min and ylimit_max
    (e.g. an RFBucket separatrix) using the numerical
    scipy.integrate.dblquad integration method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - direction: 'x' or 'y' for calculating the mean in the first
          or second argument of psi, respectively (default: 'x')
    '''

    Q = compute_zero_quad(psi, ylimit_min, ylimit_max, xmin, xmax)
    if direction is 'x':
        f = lambda y, x: x * psi(x, y)
    elif direction is 'y':
        f = lambda y, x: y * psi(x, y)
    else:
        raise ValueError('direction needs to be either "x" or "y".')

    M, error = dblquad(f, xmin, xmax, ylimit_min, ylimit_max)

    return M/Q

def compute_var_quad(psi, ylimit_min, ylimit_max, xmin, xmax, direction='x'):
    '''Compute the second moment (variance) with respect to the x or
    y direction of the distribution function psi from xmin to xmax
    between the contours ylimit_min and ylimit_max (e.g. an RFBucket
    separatrix) using the numerical scipy.integrate.dblquad integration
    method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - direction: 'x' or 'y' for calculating the standard deviation
          in the first or second argument of psi, respectively
          (default: 'x')
    '''

    Q = compute_zero_quad(psi, ylimit_min, ylimit_max, xmin, xmax)
    M = compute_mean_quad(psi, ylimit_min, ylimit_max, xmin, xmax, direction)
    if direction is 'x':
        f = lambda y, x: (x - M)**2 * psi(x, y)
    elif direction is 'y':
        f = lambda y, x: (y - M)**2 * psi(x, y)
    else:
        raise ValueError('direction needs to be either "x" or "y".')

    V, error = dblquad(f, xmin, xmax, ylimit_min, ylimit_max)

    return V/Q

def compute_cov_quad(psi, ylimit_min, ylimit_max, xmin, xmax):
    '''Compute the second moments (covariance matrix entries) of the
    distribution function psi from xmin to xmax between the contours
    ylimit_min and ylimit_max (e.g. an RFBucket separatrix) using the
    numerical scipy.integrate.dblquad integration method.
    For x the first and y the second argument of psi, return the tuple
    (variance(x), covariance(x,y), variance(y)).

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimits: contour function yielding the y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
    '''

    Q = compute_zero_quad(psi, ylimit_min, ylimit_max, xmin, xmax)
    M_x = compute_mean_quad(psi, ylimit_min, ylimit_max, xmin, xmax, 'x')
    M_y = compute_mean_quad(psi, ylimit_min, ylimit_max, xmin, xmax, 'y')

    V_x = compute_var_quad(psi, ylimit_min, ylimit_max, xmin, xmax, 'x')
    V_y = compute_var_quad(psi, ylimit_min, ylimit_max, xmin, xmax, 'y')

    f = lambda y, x: (x - M_x) * (y - M_y) * psi(x, y)
    C_xy, error = dblquad(f, xmin, xmax, ylimit_min, ylimit_max)

    C_xy /= Q

    return V_x, C_xy, V_y

### scipy.integrate.cumtrapz moment integration:

def compute_zero_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                          n_samples=513):
    '''Compute the zeroth moment of the distribution function psi from
    xmin to xmax between the contours ylimit_min and ylimit_max
    (e.g. an RFBucket separatrix) using the numerical
    scipy.integrate.cumtrapz integration method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - n_samples: integer number of sampling points for integration
    '''

    x_arr = np.linspace(xmin, xmax, num=n_samples)
    dx = x_arr[1] - x_arr[0]

    Q = 0
    for x in x_arr:
        y = np.linspace(ylimit_min(x), ylimit_max(x), num=n_samples)
        z = psi(x, y)
        Q += cumtrapz(z, y)[-1]
    Q *= dx

    return Q

def compute_mean_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                          direction='x', n_samples=513):
    '''Compute the first moment of the distribution function psi from
    xmin to xmax between the contours ylimit_min and ylimit_max
    (e.g. an RFBucket separatrix) using the numerical
    scipy.integrate.cumtrapz integration method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - direction: 'x' or 'y' for calculating the mean in the first
          or second argument of psi, respectively (default: 'x')
        - n_samples: integer number of sampling points for integration
    '''

    Q = compute_zero_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                              n_samples)

    x_arr = np.linspace(xmin, xmax, num=n_samples)
    dx = x_arr[1] - x_arr[0]

    if direction is 'x':
        f = lambda x, y: x * psi(x, y)
    elif direction is 'y':
        f = lambda x, y: y * psi(x, y)
    else:
        raise ValueError('direction needs to be either "x" or "y".')

    M = 0
    for x in x_arr:
        y = np.linspace(ylimit_min(x), ylimit_max(x), num=n_samples)
        z = f(x, y)
        M += cumtrapz(z, y)[-1]
    M *= dx

    return M/Q

def compute_var_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                         direction='x', n_samples=513):
    '''Compute the second moment (variance) with respect to the x or
    y direction of the distribution function psi from xmin to xmax
    between the contours ylimit_min and ylimit_max
    (e.g. an RFBucket separatrix) using the numerical
    scipy.integrate.cumtrapz integration method.

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - direction: 'x' or 'y' for calculating the mean in the first
          or second argument of psi, respectively (default: 'x')
        - n_samples: integer number of sampling points for integration
    '''

    Q = compute_zero_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                              n_samples)
    M = compute_mean_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                              direction, n_samples)

    x_arr = np.linspace(xmin, xmax, num=n_samples)
    dx = x_arr[1] - x_arr[0]

    if direction is 'x':
        f = lambda x, y: (x - M)**2 * psi(x, y)
    elif direction is 'y':
        f = lambda x, y: (y - M)**2 * psi(x, y)
    else:
        raise ValueError('direction needs to be either "x" or "y".')

    V = 0
    for x in x_arr:
        y = np.linspace(ylimit_min(x), ylimit_max(x), num=n_samples)
        z = f(x, y)
        V += cumtrapz(z, y)[-1]
    V *= dx

    return V/Q

def compute_cov_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax,
                         n_samples=513):
    '''Compute the second moments (covariance matrix entries) of the
    distribution function psi from xmin to xmax between the contours
    ylimit_min and ylimit_max (e.g. an RFBucket separatrix) using the
    numerical scipy.integrate.cumtrapz integration method.
    For x the first and y the second argument of psi, return the tuple
    (variance(x), covariance(x,y), variance(y)).

    Arguments:
        - psi: 2D distribution function with two arguments x and y
        - ylimit_min, ylimit_max: contour functions yielding the lower
          and the upper y(x) limit for given x
        - xmin, xmax: lower and upper limit in the first argument of psi
        - n_samples: integer number of sampling points for integration
    '''


    Q = compute_zero_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax)
    M_x = compute_mean_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax, 'x')
    M_y = compute_mean_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax, 'y')

    V_x = compute_var_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax, 'x')
    V_y = compute_var_cumtrapz(psi, ylimit_min, ylimit_max, xmin, xmax, 'y')

    x_arr = np.linspace(xmin, xmax, num=n_samples)
    dx = x_arr[1] - x_arr[0]

    C_xy = 0
    for x in x_arr:
        y = np.linspace(ylimit_min(x), ylimit_max(x), num=n_samples)
        z = (x - M_x) * (y - M_y) * psi(x, y)
        C_xy += cumtrapz(z, y)[-1]
    C_xy *= dx

    C_xy /= Q

    return V_x, C_xy, V_y


### scipy.integrate.romberg standard deviation integration:
### ==> not yet adapted to above quad and cumtrapz approaches,
###     works as Kevin had previously defined it in the RFBucketMatcher
###     (slightly corrected to make this function work here)

def compute_std_romberg(psi, ylimits, xmin, xmax, n_samples=513):
    '''
    Compute the standard deviation of the distribution function psi
    from xmin to xmax along the contours ylimits using numerical
    integration methods.
    '''

    x_arr = np.linspace(xmin, xmax, num=n_samples)
    dx = x_arr[1] - x_arr[0]

    Q, V = 0, 0
    for x in x_arr:
        y = np.linspace(0, ylimits(x), num=n_samples)
        dy = y[1] - y[0]
        z = psi(x, y)
        Q += romb(z, dy)
        z = x**2 * psi(x, y)
        V += romb(z, dy)
    Q *= dx
    V *= dx

    return np.sqrt(V/Q)



import numpy as np
from .polar import generate_2D_uniform_circular_sector

def generate_2D_pencil(num_particles, pos_cut_sigmas, dr_sigmas,
                       side='+'):

    assert side == '+' or side == '-' or side == '+-'

    if side == '+-':
        n_plus = int(num_particles/2)
        n_minus = num_particles - n_plus
        x_plus, px_plus, r_plus, theta_plus = generate_2D_pencil(n_plus,
                                                 pos_cut_sigmas, dr_sigmas, side='+')
        x_minus, px_minus, r_minus, theta_minus = generate_2D_pencil(n_minus,
                                                 pos_cut_sigmas, dr_sigmas, side='-')
        x_norm = np.concatenate([x_minus, x_plus])
        px_norm = np.concatenate([px_minus, px_plus])
        r_points = np.concatenate([r_minus, r_plus])
        theta_points = np.concatenate([theta_minus, theta_plus])

        return x_norm, px_norm, r_points, theta_points

    else:

        r_min = np.abs(pos_cut_sigmas)
        r_max = r_min + dr_sigmas
        theta_max = np.arccos(pos_cut_sigmas/(np.abs(pos_cut_sigmas) + dr_sigmas))
        target_area = r_max**2/2*(2 * theta_max - np.sin(2 * theta_max))
        generated_area = (r_max**2 - r_min**2) * theta_max

        n_gen = int(num_particles*generated_area/target_area)

        x_gen, px_gen, r_points, theta_points = generate_2D_uniform_circular_sector(
                # Should avoid regen most of the times
                num_particles=int(n_gen*1.5+100),
                r_range=(r_min, r_max),
                theta_range=(-theta_max, theta_max))

        mask = x_gen > r_min
        x_norm = x_gen[mask]
        px_norm = px_gen[mask]
        r_points = r_points[mask]
        theta_points = theta_points[mask]

        assert len(x_norm) > num_particles

        x_norm = x_norm[:num_particles]
        px_norm = px_norm[:num_particles]
        r_points = r_points[:num_particles]
        theta_points = theta_points[:num_particles]

        if side == '-':
            x_norm = -x_norm

        return x_norm, px_norm, r_points, theta_points
